/*****************************************************************************
 *
 * FILE:	alloc.c
 * DESCRIPTION:	memory allocation manager
 * DATE:	Tue, Oct 31 2000
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2000-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 2000-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: alloc.c,v 1.16.2.1 2002/01/30 13:26:33 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <string.h>
#ifndef HAVE_MEMCPY
#define	memcpy(d, s, n)		bcopy((s), (d), (n))
#define	memmove(d, s, n)	bcopy((s), (d), (n))
#endif  /* HAVE_MEMCPY */
#include <errno.h>
#include <limits.h>
#if __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	DELTA		32	/* difference between size and block */
#define	MEM_MAGIC	0xa5	/* magic number of memory block */

union align {	/* for the longest relevant alignment */
  char *	cp;
  void		(*func)(void);
  FILE *	fp;
  long		l;
  double	d;
};

typedef union _mem_block {	/* memory block */
  union	align	x;	/* never used ... */

  struct {		/* actual used ... */
    unsigned char	magic;
    unsigned int	size;
    bool_t		available;
#if	DEBUG_ALLOC
    int			no;
#endif	/* DEBUG_ALLOC */
    char *		actualBlock;
    union _mem_block *	next;
  } h;
} MemBlock;

struct _ios {
  unsigned char *	ptr;	/* current position in buffer */
  int			left;	/* left buffer size */
  unsigned char *	base;	/* buffer */
  int			size;	/* buffer size */
};


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static MemBlock *	allocateMemoryBlock(unsigned int);
static int		print(struct _ios *, unsigned char *, int);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static MemBlock *	blockChain	= NULL;
#if	DEBUG_ALLOC
static unsigned int	blockNumber	= 0;	/* number of blocks */
static unsigned int	blockTotal	= 0;	/* total size of blocks */
static unsigned int	released	= 0;	/* total free blocks */
static unsigned int	recycled	= 0;	/* total recycled blocks */
#endif	/* DEBUG */


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	allocateMemoryBlock
 * Description:		allocate a completely new block using malloc().
 * Arguments:		memory size
 * Return Value:	memory block
 */
static MemBlock *
allocateMemoryBlock(size)
	unsigned int	size;
{
  MemBlock *	block;

  block = (MemBlock *)malloc(size + sizeof(MemBlock));
  if (block == NULL) {
    wvSyslog(LOG_ERR, ERR_OUT_OF_MEMORY);
    return NULL;
  }
  block->h.magic	= MEM_MAGIC;
  block->h.size		= size;
  block->h.available	= FALSE;
  block->h.actualBlock	= (char *)(block + 1);
  block->h.next		= blockChain;

#if	DEBUG_ALLOC
  block->h.no = blockNumber++;
  blockTotal += block->h.size;
  fprintf(stderr, "[alloc]: #%04d new allocated = %3u (%uB)\n",
	  block->h.no, block->h.size, blockTotal);
#endif	/* DEBUG_ALLOC */

  return (blockChain = block);
}

#if	PARANOIA		/* for free-on-exit */
/*
 * Function Name:	cleanupMemoryBlock
 * Description:		release allocated memory blocks
 * Arguments:		none
 * Return Value:	none
 */
void
cleanupMemoryBlock(void)
{
  register MemBlock *	mbp;
  register MemBlock *	tmp;

  for (mbp = blockChain; mbp != NULL; mbp = tmp) {
    tmp = mbp->h.next;
    free((MemBlock *)mbp);
  }
}
#endif	/* PARANOIA */


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	xmalloc
 * Description:		allocate new memory space
 * Arguments:		memory size
 * Return Value:	new memory area
 */
void *
xmalloc(size)
	size_t	size;
{
  register MemBlock *	mbp;

  /* search an available memory block */
  for (mbp = blockChain; mbp != NULL; mbp = mbp->h.next) {
    if (mbp->h.available) {
      if (size <= mbp->h.size && mbp->h.size - size < DELTA) {
	mbp->h.available = FALSE;
#if	DEBUG_ALLOC
	recycled++;
	fprintf(stderr, "[alloc]: #%04d recycled[%d] = %3u (%uB) [%.2f%%]\n",
		mbp->h.no, size, mbp->h.size, blockTotal,
		(double)recycled * 100 / (double)released);
#endif	/* DEBUG_ALLOC */
	return mbp->h.actualBlock;
      }
    }
  }

  /* failed to find free memory block */
  mbp = allocateMemoryBlock(size);

  return (mbp == NULL ? NULL : mbp->h.actualBlock);
}

/*
 * Function Name:	xfree
 * Description:		release allocated memory space
 * Arguments:		memory area
 * Return Value:	none
 */
void
xfree(ptr)
	void *	ptr;
{
  register MemBlock *	mbp;

  /* check invalid pointer */
  if (ptr == NULL) {
    return;
  }

  mbp = (MemBlock *)(ptr - sizeof(MemBlock));
  if (mbp->h.magic == MEM_MAGIC) {
    memset((char *)mbp->h.actualBlock, 0, mbp->h.size);
    mbp->h.available = TRUE;
#if	DEBUG_ALLOC
    released++;
    fprintf(stderr, "[alloc]: #%04d free memblock = %3u (%uB)\n",
	    mbp->h.no, mbp->h.size, blockTotal);
#endif	/* DEBUG_ALLOC */
  }
  else {	/* the memory of ptr is allocated by malloc(). */
    free(ptr);
#if	DEBUG_ALLOC
    fputs("[alloc]: free direct malloc()\n", stderr);
#endif	/* DEBUG_ALLOC */
  }
}

/*
 * Function Name:	xcalloc
 * Description:		allocate new memory space and clear memory area
 * Arguments:		object number, object size
 * Return Value:	new memory area
 */
void *
xcalloc(num, size)
	size_t	num;
	size_t	size;
{
  register void *	p;

  size *= num;
  p = xmalloc(size);
  if (p != NULL) {
    memset((char *)p, 0, size);
  }
  return p;
}

/*
 * Function Name:	xrealloc
 * Description:		reallocate used memory space
 * Arguments:		memory pointer, new memory size
 * Return Value:	memory pointer
 */
void *
xrealloc(ptr, size)
	void *	ptr;
	size_t	size;
{
  register MemBlock *	mbp;
  register void *	nptr;

  /* behave identically to xmalloc() */
  if (ptr == NULL) {
    return xmalloc(size);
  }

  mbp = (MemBlock *)(ptr - sizeof(MemBlock));
  if (mbp->h.magic == MEM_MAGIC) {
    nptr = xcalloc(size, sizeof(char));
    if (nptr == NULL) {
      wvSyslog(LOG_ERR, ERR_OUT_OF_MEMORY);
      return ptr;
    }
    else {
      memcpy(nptr, mbp->h.actualBlock, mbp->h.size);
      XFree(ptr);
      return nptr;
    }
  }

  return NULL;
}

#if	0
/*
 * Function Name:	xreallocf
 * Description:		reallocate used memory space with free pointer
 * Arguments:		memory pointer, new memory size
 * Return Value:	memory pointer
 */
void *
xreallocf(ptr, size)
	void *	ptr;
	size_t	size;
{
  void *	nptr;

  nptr = xrealloc(ptr, size);
  if (!nptr && ptr) {
    XFree(ptr);
  }
  return nptr;
}
#endif

/*
 * Function Name:	xstrdup
 * Description:		duplicate a string
 * Arguments:		string
 * Return Value:	new string
 */
char *
xstrdup(str)
	const char *	str;
{
  size_t	len;
  char *	copy;

  len	= strlen(str) + 1;
  copy	= xmalloc(len);
  if (copy == NULL) {
    return NULL;
  }
  memcpy(copy, str, len);
  return copy;
}

/*
 * Function Name:	xstrndup
 * Description:		duplicate a string
 * Arguments:		string, length
 * Return Value:	new string
 */
char *
xstrndup(str, len)
	const char *	str;
	size_t		len;
{
  size_t	size = len + 1;
  char *	dst;
  char *	copy;
  int		i;

  copy = dst = xmalloc(size);
  if (copy == NULL) {
    return NULL;
  }
  for (i = 0; i < len; i++) {
    *dst++ = *str++;
  }
  *dst = '\0';
  return copy;
}

/*
 * Function Name:	xasprintf
 * Description:		formatted output convertion with dynamic memory
 *			allocation
 * Arguments:		same as asprintf
 * Return Value:	successful: number of characters, otherwise: -1
 * Note:		based on FreeBSD:/usr/src/lib/libc/stdio/asprintf.c
 */
#define	LONGINT		0x001	/* long integer */
#define	HEXDIGIT	0x002	/* hex-digit */
int
#if __STDC__
xasprintf(char **str, char const *fmt, ...)
#else
xasprintf(str, fmt, va_alist)
	char **		str;
	const char *	fmt;
	va_dcl
#endif
{
  int		ret;	/* return value accumulator */
  va_list	ap;
  struct _ios	buf;
  char		ch;	/* character from fmt */
  char *	cp;	/* handy char pointer (short term usage) */
  int		n;	/* handy integer (short term usage) */
  int		d;	/* int type (short term usage ) */
  long int	l;	/* long int type (short term usage ) */
  char *	s;	/* char * type (short term usage ) */
  int		flags;

#if	__STDC__
  va_start(ap, fmt);
#else
  va_start(ap);
#endif	/* __STDC__ */
  buf.size = 128;
  buf.base = buf.ptr = (unsigned char *)xmalloc((size_t)buf.size);
  if (buf.base == NULL) {
    goto error;
  }
  buf.left = buf.size - 1;	/* Leave room for the NULL */

  ret = 0;
  /* scan the format for conversions ('%' character) */
  for (;;) {
    for (cp = fmt; (ch = *fmt) != EOL && ch != '%'; fmt++) {
    	;	/* nothing to do */
    }
    if ((n = fmt - cp) != 0) {
      if ((unsigned)ret + n > INT_MAX) {
	goto error;
      }
      if (print(&buf, cp, n) == -1) {
	goto error;
      }
      ret += n;
    }
    if (ch == EOL) {
      goto done;
    }
    fmt++;	/* skip over '%' */
    flags = 0;
    n	  = 0;
restart:
    ch = *fmt++;
    switch (ch) {
      case ' ':
	goto restart;
      case '\'':
      case '#': case '*': case '-': case '+': case '.':
      case '0':
      case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
	goto restart;
      case 'c':
	/* Note: char is promoted to int. */
	//thd->ios.buf[thd->ios.next++] = va_arg(ap, int);
	*cp = va_arg(ap, int);
	if (print(&buf, cp, 1) == -1) {
	  goto error;
	}
	continue;
      case 's':
	s = va_arg(ap, char *);
	n = strlen(s);
	break;
      case 'l':
	flags |= LONGINT;
	goto restart;
      case 'x':
	flags |= HEXDIGIT;
      case 'd':
      case 'i':
	if (flags & LONGINT) {
	  l = va_arg(ap, long int);
	}
	else {
	  d = va_arg(ap, int);
	}
	if (flags & HEXDIGIT) {
	  if (flags & LONGINT) {
	    s = itoxa((unsigned long int)l);
	  }
	  else {
	    s = itoxa((unsigned int)d);
	  }
	}
	else {
	  if (flags & LONGINT) {
	    s = itoa(l);
	  }
	  else {
	    s = itoa(d);
	  }
	}
	n = strlen(s);
	break;
      default:
	if (ch == EOL) {
	  goto done;
	}
	goto restart;
    }
    if (print(&buf, s, n) == -1) {
      goto error;
    }
    ret += n;
  }

done:
  va_end(ap);
  *buf.ptr = EOL;
  *str = (char *)buf.base;
  return ret;

error:
  va_end(ap);
  xfree(buf.base);
  *str = NULL;
  return -1;
}

/*
 * Function Name:	print
 * Description:		subroutine for xasprintf()
 * Arguments:		buffer structure, string pointer, string length
 * Return Value:	successful: number of characters, otherwise: -1
 */
static int
print(buf, ptr, len)
	struct _ios *	buf;
	unsigned char *	ptr;
	int		len;
{
  int	total = buf->size - buf->left - 1;

retry:
  if (buf->left >= len) {
    memcpy((unsigned char *)buf->ptr, (unsigned char *)ptr, (size_t)len);
    buf->left -= len;
    buf->ptr  += len;
  }
  else {
    buf->size *= 2;
    buf->base = (unsigned char *)xrealloc(buf->base, (size_t)buf->size);
    if (buf->base == NULL) {
      return -1;
    }
    buf->ptr  = buf->base + total;
    buf->left = buf->size - total - 1;	/* leave room for EOL */
    goto retry;
  }

  return len;
}

