/*****************************************************************************
 *
 * FILE:	get.c
 * DESCRIPTION:	functions for HEAD/GET method
 * DATE:	Sat, Oct  9 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: get.c,v 1.41.2.3 2002/01/29 10:24:25 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	openDirectory(Thread *);
static bool_t	safeSymbolicLink(Thread *, String);
static bool_t	checkSymbolicLink(Thread *, String);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	openRequestFile
 * Description:		open the request file
 * Arguments:		thread
 * Return Value:	none
 */
void
openRequestFile(thd)
	register Thread *	thd;
{
  /*
   * following symbolic link is restricted to secure region.
   */
  if (httpd->vhost->symlinks == FALSE &&
      safeSymbolicLink(thd, thd->path) == FALSE) {
    return;
  }

  /*
   * open file
   */
  thd->fp = fopen(thd->path, "r");
again:
  if (thd->fp == NULL) {
    switch (errno) {
      case ENOENT:
	if (thd->ctype != C_DIR) {
#if	DEBUG
	  fprintf(stderr, "[open]: Not found path = %s\n", thd->path);
#endif	/* DEBUG */
	  setResponseCode(thd, CODE_NOT_FOUND);
	}
	return;
      case EACCES:
	setResponseCode(thd, CODE_FORBIDDEN);
	return;
      default:
	setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
	return;
    }
  }

  /* CGI & SSI */
  if (thd->method.type == M_GET &&
      (thd->ctype == C_CGI || thd->ctype == C_SSI)) {
    fclose(thd->fp);
    return;
  }

  /* already set response error message */
  if (thd->ctype == C_MSG) {
    return;
  }

  /* get the file/directory status */
  if (fstat(fileno(thd->fp), &thd->file) == -1) {
    setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
    return;
  }

  /* get content */
  if (S_ISDIR(thd->file.st_mode)) {	/* directory */
    openDirectory(thd);
    goto again;
  }
  else if (S_ISREG(thd->file.st_mode)) {	/* regular file	*/
    checkConditionalRequest(thd);
    if (thd->file.st_size == 0) {
      setResponseCode(thd, CODE_NO_CONTENT);
    }
  }
  else {
    setResponseCode(thd, CODE_FORBIDDEN);
  }
  setMimeType(thd);
}

/*
 * Function Name:	openDirectory
 * Description:		open directory
 * Arguments:		thread
 * Return Value:	none
 */
static void
openDirectory(thd)
	register Thread *	thd;
{
  String	path;

  /*
   * find default index file (DefaultIndex) defined in wyvern.conf
   */
  fclose(thd->fp);	/* close old file pointer */
  thd->ctype = C_DIR;	/* directory listing mode */

  /* redirect an illegal request URI */
  if (thd->requestURI[strlen(thd->requestURI)-1] != '/') {
    if (xasprintf(&thd->response.location, "%s/", thd->requestURI) == -1) {
      setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
    }
    else {
      setResponseCode(thd, CODE_MOVED_PERMANENTLY);
    }
  }
  /* set new path with DefaultIndex in wyvern.conf */
  else if (xasprintf(&path, "%s/%s", thd->path, httpd->vhost->dirIndex) != -1) {
    /* open new file */
    thd->fp = fopen(path, "r");
    XFree(path);
  }
  else {
    setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
  }
}


/*
 * Function Name:	safeSymbolicLink
 * Description:		track a symbolic link file
 * Arguments:		thread, file path
 * Return Value:	Safe SymbolicLink: TRUE, otherwise: FALSE
 */
#define	MAX_SYMLINKS	5
static bool_t
safeSymbolicLink(thd, path)
	register Thread *	thd;
	String			path;
{
  bool_t	trusted	 = FALSE;
  static int	symlinks = 0;
  struct stat	sbuf;

  if (lstat(path, &sbuf) == 0) {
    if (S_ISLNK(sbuf.st_mode)) {	/* symbolic link */
      char	realPath[MAX_PATH_LEN];

      if (++symlinks > MAX_SYMLINKS) {
	goto error;
      }
      memset(realPath, 0, sizeof(realPath));
      if (readlink(path, realPath, sizeof(realPath) - 1) == -1) {
	goto error;
      }
      return safeSymbolicLink(thd, realPath);
    }
    trusted = checkSymbolicLink(thd, path);
  }
  else if (path[0] != '/' && path[0] != '.' &&
	   path[1] != '/' && path[1] != '.') {
#if	DEBUG
    fprintf(stderr, "[safe] path = %s\n", path);
#endif	/* DEBUG */
    trusted = TRUE;
  }
  else if (symlinks == 0 && errno == ENOENT) {
    setResponseCode(thd, CODE_NOT_FOUND);
  }
  else {
    setResponseCode(thd, CODE_FORBIDDEN);
  }

  if (symlinks > 0) {
    symlinks = 0;
    wvSyslog(trusted ? LOG_NOTICE : LOG_ALERT,
	     "request URI '%s' is %s symbolic file '%s'.\n",
	     thd->requestURI,
	     trusted ? "a" : "an unsecure",
	     path);
  }

  return trusted;

error:
  setResponseCode(thd, CODE_FORBIDDEN);
  return FALSE;
}

/*
 * Function Name:	checkSymbolicLink
 * Description:		check a symbolic link file
 * Arguments:		thread, file path
 * Return Value:	Successful: TRUE, otherwise: FALSE
 */
static bool_t
checkSymbolicLink(thd, path)
	register Thread *	thd;
	String			path;
{
  register unsigned int	i;
  bool_t		trusted = FALSE;

  for (i = 0; i < ALIAS_TABLE_SIZE && trusted == FALSE; i++) {
    register Alias *	ap;

    for (ap = httpd->vhost->aliasTable[i]; ap != NULL; ap = ap->next) {
      if (ap->type == A_REDIRECT) {
       	continue;
      }
      /* path exists in Alias table or in DocumentRoot */
      if (memcmp(path, ap->realName, ap->realLen) == 0 ||
	  memcmp(path, httpd->vhost->docRoot,
		 strlen(httpd->vhost->docRoot)) == 0) {
	trusted = TRUE;
	break;
      }
    }
  }

  if (trusted == FALSE) {
    if (memcmp(path, httpd->serverRoot, strlen(httpd->serverRoot)) == 0 ||
     	(thd->userDir != NULL &&
	 memcmp(path, thd->userDir, strlen(thd->userDir)) == 0) ||
      	(path[0] != '/' && memcmp(path, "../", 3) != 0)) {
      trusted = TRUE;
    }
    else {
      setResponseCode(thd, CODE_FORBIDDEN);
    }
  }

  return trusted;
}

