/*****************************************************************************
 *
 * FILE:	mod_cgi.c
 * DESCRIPTION:	CGI module for wyvern
 * DATE:	Sat, Sep 16 2000
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2000-2002 졿Kouichi ABE (WALL), All rights reserved.
 * REFERENCE:	draft-coar-cgi-v11-03.txt
 * LICENSE:
 *  Copyright (C) 2000-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_cgi.c,v 1.23.2.5 2002/01/30 13:26:40 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdlib.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_VFORK_H
#include <vfork.h>
#endif	/* HAVE_VFORK_H */
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/wait.h>
#include <signal.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	declaration of global functions
 *
 *****************************************************************************/
extern void	runScript(Thread *);


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	trapCGI(int);
static void	waitChild(void);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	_init
 * Description:		set SIGCHLD trap
 * Arguments:		none
 * Return Value:	none
 */
void
_init(void)
{
  struct sigaction	act;

  sigemptyset(&act.sa_mask);
  act.sa_flags	 = SA_NOCLDSTOP;
  act.sa_handler = (void *)waitChild;
  sigaction(SIGCHLD, &act, NULL);
#if	DEBUG
  fputs("[mod_cgi]: plugged-in\n", stderr);
#endif	/* DEBUG */
}

/*
 * Function Name:	_fini
 * Description:		unset SIGCHLD trap
 * Arguments:		none
 * Return Value:	none
 */
void
_fini(void)
{
  struct sigaction	act;

  sigemptyset(&act.sa_mask);
  act.sa_flags	 = 0;
  act.sa_handler = SIG_IGN;
  sigaction(SIGCHLD, &act, NULL);
}


/*****************************************************************************
 *****************************************************************************/

/*
 * SERVER_SOFTWARE:	the name and version of the server software
 * SERVER_NAME:		the name of the server
 * GATEWAY_INTERFACE:	the revision of the CGI specification
 *
 * SERVER_PROTOCOL:	the protocol version of the HTTP
 * SERVER_PORT:		the port which the request is received on the server
 *
 * REQUEST_METHOD:	the method of the HTTP
 * SCRIPT_NAME:		the identity of the CGI script
 * PATH_INFO:		the path to be interpreted by the CGI script
 * PATH_TRANSLATED:	the location in the server's document
 * QUERY_STRING:	a URL-encoded string
 * REMOTE_HOST:		the FQDN of the client
 * REMOTE_ADDR:		the IP address of the client
 * REMOTE_PORT:		the remote port of the client
 * AUTH_TYPE:		the requests "Authorization"
 * REMOTE_USER:		the remote user-ID
 * REMOTE_IDENT:	the identity information reported about the connection
 *			by a RFC 1413
 * CONTENT_TYPE:	the Internet Media Type
 * CONTENT_LENGTH:	the size of the message-body entity
 *
 *
 * HTTP_ACCEPT:
 * HTTP_ACCEPT_ENCODING:
 * HTTP_ACCEPT_LANGUAGE:
 * HTTP_USER_AGENT:
 * HTTP_REFERER:
 *
 * DOCUMENT_ROOT:	NCSA and APACHE added -- not in CGI spec
 * SERVER_ROOT:		NCSA added
 * SERVER_ADMIN:	APACHE added
 */

/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	runScript
 * Description:		execute CGI script
 * Arguments:		thread
 * Return Value:	none
 */
#define	READ_FD		0	/* descriptor to read */
#define	WRITE_FD	1	/* descriptor to write */
void
runScript(thd)
	register Thread *	thd;
{
  char		cwd[MAX_PATH_LEN];	/* CGI working directory */
  String	argv[2] = { NULL, NULL };
  int		pfd[2];
  pid_t		pid;

  /* set CGI working directory */
  memset(cwd, 0, sizeof(cwd));
  if (xstrncpy(cwd, thd->path, sizeof(cwd)) != NULL &&
      /* create CGI environments */
      setScriptEnvironment(thd) != FALSE &&
      /* make pipeline between parent process and child process */
      pipe(pfd) != -1) {
    String	p = strrchr(cwd, '/');
    if (p != NULL) {
      *p = EOL;
    }
  }
  else {
    setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
    return;
  }

  /*
   * execute a CGI script
   */
  pid = vfork();
  switch (pid) {
    case -1:	/* fork error */
      setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
      return;

    case 0:	/* child process */
      xclose(pfd[READ_FD]);
      dup2(pfd[WRITE_FD], STDOUT_FILENO);
      dup2(fileno(httpd->errorLog), STDERR_FILENO);
      xclose(pfd[WRITE_FD]);
      if (thd->method.type == M_POST) {
	dup2(thd->cgi.fd, STDIN_FILENO);	/* bind cfd to stdin */
      }
      xclose(thd->sd);			/* close socket descriptor */

      argv[0] = thd->path;
      if (chdir(cwd) == -1 ||
	  execve(thd->path, argv, thd->cgi.env) == -1) {
	trapCGI(errno);
      }
      _exit(CGI_ERR_STATUS);

    default:	/* parent process */
      thd->cgi.pid = pid;
      thd->state   = CGI_WAITING;
      xclose(pfd[WRITE_FD]);
      thd->fp = fdopen(dup(pfd[READ_FD]), "r");
      xclose(pfd[READ_FD]);
#if	DEBUG
      fprintf(stderr, "[mod_cgi]: pid = %d\n", thd->cgi.pid);
#endif	/* DEBUG */
      break;
  }
}

/*
 * Function Name:	trapCGI
 * Description:		send error message of CGI
 * Arguments:		error number
 * Return Value:	none
 */
static void
trapCGI(e)
	int	e;
{
#if	USE_JP_MESG
#define	ERR_CGI_TITLE	"CGI 顼"
#define	ERR_ACCES_H1	"CGI ץ˼¹ԵĤޤ"
#define	ERR_ACCES_MESG	"ꤵ줿 CGI ץ˼¹ԵĤޤ"
#define	ERR_OTHER_H1	"¾Υ顼"
#define	ERR_OTHER_MESG	"CGI ¹Իͽ̥顼ȯޤ"
#else
#define	ERR_CGI_TITLE	"CGI Error"
#define	ERR_ACCES_H1	"Permission denied"
#define	ERR_ACCES_MESG	"The CGI file has no executable permission."
#define	ERR_OTHER_H1	"Unexcepted error"
#define	ERR_OTHER_MESG	"Server had unexpected error."
#endif	/* USE_JP_MESG */
  fputs("Content-Type: text/html" CRLF CRLF, stdout);
  fprintf(stdout,
	  "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">\n"
	  "<html>\n"
	  "<head>\n"
	  "<title>%s</title>\n"
	  "</head>\n"
	  "<body bgcolor=\"white\">\n",
	  ERR_CGI_TITLE);

  fputs("<h1 align=\"center\">", stdout);
  switch (e) {
    case EACCES:
      fprintf(stdout, "%s</h1>\n<p align=\"center\">%s",
	      ERR_ACCES_H1, ERR_ACCES_MESG);
      break;
    default:
      fprintf(stdout, "%s</h1>\n<p align=\"center\">%s",
	      ERR_OTHER_H1, ERR_OTHER_MESG);
      break;
  }
  fprintf(stdout,
	  "</p>\n"
	  "<hr>\n"
	  "<address>\n"
	  "Powered by&nbsp;"
	  "<a href=\"%s\">%s</a>\n"
	  "&nbsp;&lt;<a href=\"mailto:%s\">%s</a>&gt;\n"
	  "</address>\n"
	  "</body>\n"
	  "</html>\n",
	  SERVER_ADDRESS, SERVER_VERSION,
	  httpd->vhost->serverAdmin, httpd->vhost->serverAdmin);
  fflush(stdout);
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	waitChild
 * Description:		wait for child process termination
 * Arguments:		none
 * Return Value:	none
 */
static void
waitChild(void)
{
  sigset_t	set;
  sigset_t	oset;
  pid_t		pid;
  int		status;

  sigprocmask(SIG_SETMASK, NULL, &set);	/* get current signal mask */
  sigaddset(&set, SIGCHLD);		/* add SIGCHLD */
  sigprocmask(SIG_BLOCK, &set, &oset);	/* block SIGCHLD */

  while ((pid = waitpid((pid_t)0, &status, WNOHANG)) > 0) {
    register Thread *	thd;
    for (thd = getThreadHead(); thd != NULL; thd = thd->next) {
      if (pid == thd->cgi.pid) {
	if (thd->fp != NULL) {
	  char		buf[STD_BUFSIZE];
	  struct stat	sbuf;

	  /* get file status if possible */
	  if (fstat(fileno(thd->fp), &sbuf) == 0) {
	    thd->file = sbuf;
	  }

	  /* CGI program output 'Status' line? */
	  memset(buf, 0, sizeof(buf));
	  while (fgets(buf, sizeof(buf), thd->fp) != NULL) {
	    /* end of response header */
	    if (buf[0] == EOL || buf[0] == CR || buf[0] == LF) {
	      break;
	    }
	    /* CGI sends Content-Type line in response line */
	    else if (strncasecmp(buf, "CONTENT-TYPE:", 13) == 0) {
	      thd->mtype = thd->response.content_type = getAnyHeaderField(buf);
	    }
	    /* CGI sends Content-Length line in response line */
	    else if (strncasecmp(buf, "CONTENT-LENGTH:", 15) == 0) {
	      thd->file.st_size = thd->response.content_length = atoi(&buf[15]);
	    }
	    /* 'Status' line */
	    else if (strncasecmp(buf, "STATUS:", 7) == 0) {
	      thd->cgi.status = getAnyHeaderField(buf);
	    }
	    else if (strncasecmp(buf, "HTTP/1.1", 8) == 0 ||
		     strncasecmp(buf, "HTTP/1.0", 8) == 0 ||
		     strncasecmp(buf, "HTTP/0.9", 8) == 0) {
	      /* invalid output from CGI program */
	    }
	    /* another response headers */
	    else {
	      StrList *	header = (StrList *)xcalloc(1, sizeof(StrList));
	      if (header == NULL) {
		wvSyslog(LOG_ERR, ERR_OUT_OF_MEMORY);
		continue;
	      }
	      header->text	= xstrdup(buf);
	      header->next	= thd->cgi.headers;
	      thd->cgi.headers	= header;
	    }
	    /* clear buffer */
	    memset(buf, 0, sizeof(buf));
	  }
	}
#if	DEBUG
	fprintf(stderr, "[SIGNAL] pid=%d, status=%d, size=%d\n",
		thd->cgi.pid, WEXITSTATUS(status), (int)thd->file.st_size);
#endif	/* DEBUG */
	thd->cgi.pid = -1;	/* cgi process completed */
	thd->state   = SEND_READY;
	break;
      }
    } /* for */
  }

  sigprocmask(SIG_SETMASK, &oset, NULL);	/* reset signal mask */
}

