/*****************************************************************************
 *
 * FILE:	mod_gzip.c
 * DESCRIPTION:	gzip module using the zlib compression library
 * DATE:	Thu, May 17 2001
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2001-2002 졿Kouichi ABE (WALL), All rights reserved.
 * NOTE:	Based on minigzip.c
 * LICENSE:
 *  Copyright (C) 2001-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_gzip.c,v 1.15.2.2 2002/01/27 16:10:30 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#if	HAVE_MMAP
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#endif	/* HAVE_MMAP */
#include "wyvern.h"
#include <zlib.h>	/* this order is important. */
#include "functions.h"


/*****************************************************************************
 *
 *	declaration of global functions
 *
 *****************************************************************************/
extern void	gzipCompress(Thread *);


/*****************************************************************************
 *
 *	definition of macros
 *
 *****************************************************************************/
#define	GZIP_MODE	"wb6"	/* See zlib.h */
#define GZIP_BUFSIZE	(HUGE_BUFSIZE*4)
#define	MIN_FILE_SIZE	400


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
#if	HAVE_MMAP
static int	gzipCompressWithMmap(FILE *, gzFile);
#endif	/* HAVE_MMAP */
static void	reopen(Thread *, String);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

#if	DEBUG
void
_init(void)
{
  fputs("[mod_gzip]: plugged-in\n", stderr);
}
#endif	/* DEBUG */

/*
 * Function Name:	gzipCompress
 * Description:		compress content file with gzip compress
 * Arguments:		thread
 * Return Value:	none
 */
void
gzipCompress(thd)
	register Thread *	thd;
{
  String		gzfile;
  register FILE *	in = thd->fp;
  gzFile		out;
  static char		buf[GZIP_BUFSIZE];
  bool_t		flag;	/* successful: TRUE, otherwise: FALSE */

#if	DEBUG
   fprintf(stderr, "[mod_gzip] mime-type: %s\n", thd->mtype);

  /*
   * no gzip when content size is less than MIN_FILE_SIZE bytes
   * or partial content request.
   */
  if (thd->file.st_size < MIN_FILE_SIZE || thd->ctype == C_PAR) {
    fprintf(stderr, "[mod_gzip] too small or partial request, no gzip: %d/%d\n",
	    (int)thd->file.st_size, MIN_FILE_SIZE);
    return;
  }

  /* no supported by client */
  if (thd->request.accept_encoding == NULL ||
      strstr(thd->request.accept_encoding, "gzip") == NULL ||
      thd->mtype == NULL ||
      strncmp(thd->mtype, "text/", 5)  != 0) {
    fputs("[mod_gzip] no acceptable gzip encoding\n", stderr);
    return;
  }
#else
  /*
   * no gzip when content size is less than MIN_FILE_SIZE bytes
   * or partial content request.
   */
  if (thd->file.st_size < MIN_FILE_SIZE ||/* content size < MIN_FILE_SIZE */
      thd->ctype == C_PAR ||		/* partial content */
      /* no supported by client */
      thd->request.accept_encoding == NULL ||
      strstr(thd->request.accept_encoding, "gzip") == NULL ||
      /* no MIME type */
      thd->mtype == NULL  ||
      strncmp(thd->mtype, "text/", 5)  != 0) {
    return;
  }
#endif	/* DEBUG */

  /* open gzip'ed output file */
  if (xasprintf(&gzfile, "%s/gzfile%u",
		httpd->tmpDir, (unsigned int)thd) == -1) {
    return;
  }
  out = gzopen(gzfile, GZIP_MODE);
  if (out == NULL) {
#if	DEBUG
    fputs("[mod_gzip] failed to open working file\n", stderr);
#endif	/* DEBUG */
    goto done;
  }

#if	HAVE_MMAP
  /*
   * Try first compressing with mmap.
   * If mmap fails, use the normal fread loop.
   */
  if (gzipCompressWithMmap(in, out) == Z_OK) {
    reopen(thd, gzfile);
#if	DEBUG
    fputs("[mod_gzip] successful with mmap\n", stderr);
#endif	/* DEBUG */
    goto done;
  }
#endif	/* HAVE_MMAP */

  flag = TRUE;
  memset(buf, 0, sizeof(buf));
  while (!feof(in) && !ferror(in)) {
    size_t	len;

    len = fread(buf, 1, sizeof(buf), in);
    if (gzwrite(out, buf, (unsigned)len) != len) {
      flag = FALSE;
      break;
    }
  }
  if (gzclose(out) != Z_OK) {	/* close the compressed content file */
    flag = FALSE;
  }

  if (flag) {	/* successful */
    reopen(thd, gzfile);
#if	DEBUG
    fputs("[mod_gzip] successful\n", stderr);
#endif	/* DEBUG */
  }
  else {
    (void)unlink(gzfile);	/* remove temporary file */
    (void)fseek(thd->fp, 0L, SEEK_SET);	/* rewind */
  }

done:
  xfree((String)gzfile);
}

#if	HAVE_MMAP
/*
 * Function Name:	gzipCompressWithMmap
 * Description:		compress content file with gzip compress using mmap
 * Arguments:		file pointer, gzip'ed file pointer
 * Return Value:	successful: Z_OK, otherwise: Z_ERRNO
 */
static int
gzipCompressWithMmap(in, out)
	FILE *	in;
	gzFile	out;
{
  int		fd = fileno(in);
  struct stat	sbuf;
  caddr_t	buf;	/* mmap'ed buffer for the entire input file */
  off_t		size;	/* length of the input file */
  size_t	len;
  bool_t	flag;	/* successful: TRUE, otherwise: FALSE */

  /* Determine the size of the file, needed for mmap: */
  if (fstat(fd, &sbuf) == -1) {
    return Z_ERRNO;
  }

  size = sbuf.st_size;
  if (size <= 0) {
    return Z_ERRNO;
  }

  /* Now do the actual mmap: */
  buf = mmap((caddr_t) 0, (size_t)size, PROT_READ, MAP_SHARED, fd, (off_t)0); 
  if (buf == (caddr_t)(-1)) {
    return Z_ERRNO;
  }
  flag = TRUE;

  /* Compress the whole file at once: */
  len = gzwrite(out, (String)buf, (unsigned)size);
  if (len != (int)size) {
    flag = FALSE;
  }
  munmap(buf, (size_t)size);

  if (gzclose(out) != Z_OK) {
    flag = FALSE;
  }

  return (flag ? Z_OK : Z_ERRNO);
}
#endif /* HAVE_MMAP */


/*
 * Function Name:	reopen
 * Description:		reopen file of content
 * Arguments:		thread, gzip'ed file
 * Return Value:	none
 */
static void
reopen(thd, gzfile)
	Thread *	thd;
	String		gzfile;
{
  register FILE *	fin;
  struct stat		sbuf;

  /* open new file */
  fin = fopen(gzfile, "r");
  if (fin == NULL) {
    (void)fseek(thd->fp, 0L, SEEK_SET);	/* rewind */
    return;
  }

  /* get actual content status */
  if (fstat(fileno(fin), &sbuf) == -1) {
    (void)fclose(fin);
    (void)fseek(thd->fp, 0L, SEEK_SET);	/* rewind */
    return;
  }

  /* the file will be removed when it's closed. */
  (void)unlink(gzfile);

  /* change file descriptor from old to new */
  (void)fclose(thd->fp);
  thd->fp = fin;
  thd->gzippct = (1.0 - (double)sbuf.st_size / (double)thd->file.st_size) * 100;
  thd->file.st_size = sbuf.st_size;
  thd->response.content_encoding = xstrdup("gzip");
}

