/*****************************************************************************
 *
 * FILE:	script.c
 * DESCRIPTION:	execute CGI script
 * DATE:	Sun, Sep 26 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2002 졿Kouichi ABE (WALL), All rights reserved.
 * REFERENCE:	draft-coar-cgi-v11-03.txt
 * LICENSE:
 *   Copyright (C) 1999-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: script.c,v 1.42.2.2 2002/01/27 16:09:19 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdlib.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/wait.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	initServerEnvironment
 * Description:		initialize server environment for CGI/SSI
 * Arguments:		none
 * Return Value:	none
 * Notes:
 *
 * SERVER_SOFTWARE:	the name and version of the server software
 * SERVER_NAME:		the name of the server
 * GATEWAY_INTERFACE:	the revision of the CGI specification
 *
 * SERVER_PROTOCOL:	the protocol version of the HTTP
 * SERVER_PORT:		the port which the request is received on the server
 *
 * REQUEST_METHOD:	the method of the HTTP
 * SCRIPT_NAME:		the identity of the CGI script
 * PATH_INFO:		the path to be interpreted by the CGI script
 * PATH_TRANSLATED:	the location in the server's document
 * QUERY_STRING:	a URL-encoded string
 * REMOTE_HOST:		the FQDN of the client
 * REMOTE_ADDR:		the IP address of the client
 * REMOTE_PORT:		the remote port of the client
 * AUTH_TYPE:		the requests "Authorization"
 * REMOTE_USER:		the remote user-ID
 * REMOTE_IDENT:	the identity information reported about the connection
 *			by a RFC 1413
 * CONTENT_TYPE:	the Internet Media Type
 * CONTENT_LENGTH:	the size of the message-body entity
 *
 *
 * HTTP_ACCEPT:
 * HTTP_ACCEPT_ENCODING:
 * HTTP_ACCEPT_LANGUAGE:
 * HTTP_USER_AGENT:
 * HTTP_REFERER:
 *
 * DOCUMENT_ROOT:	NCSA and APACHE added -- not in CGI spec
 * SERVER_ROOT:		NCSA added
 * SERVER_ADMIN:	APACHE added
 */
void
initServerEnvironment(void)
{
  /* allocate memory area of environment variables */
  httpd->env = (String *)xcalloc(MAX_CGI_ENVNO, sizeof(String *));
  if (httpd->env == NULL) {
    die(ERR_OUT_OF_MEMORY);
  }
  httpd->envno = 0;

  /*
   * set common environment variables, but ignore errors.
   */
  xasprintf(&httpd->env[httpd->envno++], "PATH=%s", DEFAULT_PATH);
  xasprintf(&httpd->env[httpd->envno++], "SERVER_SOFTWARE=%s", SERVER_VERSION);
  xasprintf(&httpd->env[httpd->envno++], "GATEWAY_INTERFACE=%s", CGI_VERSION);
  xasprintf(&httpd->env[httpd->envno++], "SERVER_PROTOCOL=%s", HTTP_VERSION);
  xasprintf(&httpd->env[httpd->envno++], "SERVER_PORT=%d", httpd->port);
  xasprintf(&httpd->env[httpd->envno++], "SERVER_ROOT=%s", httpd->serverRoot);
}

#if	PARANOIA	/* for free-on-exit */
/*
 * Function Name:	freeServerEnvironment
 * Description:		clear server environment variables
 * Arguments:		none
 * Return Value:	none
 */
void
freeServerEnvironment(void)
{
  int	i;

  for (i = httpd->envno - 1; i >= 0; i--) {
    XFree(httpd->env[i]);
  }
  XFree(httpd->env);
}
#endif	/* PARANOIA */


/*
 * Function Name:	setScriptEnvironment
 * Description:		set script environments for CGI/SSI
 * Arguments:		Thread
 * Return Value:	Successful: TRUE, otherwise: FALSE
 */
bool_t
setScriptEnvironment(thd)
	register Thread *	thd;
{
  /*
   * create CGI environments
   */
  thd->cgi.env = (String *)xcalloc(MAX_CGI_ENVNO, sizeof(String *));
  if (thd->cgi.env == NULL) {
    setResponseCode(thd, CODE_INTERNAL_SERVER_ERROR);
    wvSyslog(LOG_ERR, ERR_OUT_OF_MEMORY);
    return FALSE;
  }

  for (thd->cgi.envno = 0; thd->cgi.envno < httpd->envno; thd->cgi.envno++) {
    thd->cgi.env[thd->cgi.envno] = httpd->env[thd->cgi.envno];
  }
  xasprintf(&thd->cgi.env[thd->cgi.envno++], "SERVER_NAME=%s",
	    httpd->vhost->serverName);
  xasprintf(&thd->cgi.env[thd->cgi.envno++], "SERVER_ADMIN=%s",
	    httpd->vhost->serverAdmin);
  xasprintf(&thd->cgi.env[thd->cgi.envno++], "DOCUMENT_ROOT=%s",
	    httpd->vhost->docRoot);
  xasprintf(&thd->cgi.env[thd->cgi.envno++], "REQUEST_METHOD=%s",
	    thd->method.name);
  xasprintf(&thd->cgi.env[thd->cgi.envno++], "SCRIPT_NAME=%s", thd->requestURI);
  if (thd->cgi.path_info != NULL) {
    xasprintf(&thd->cgi.env[thd->cgi.envno++], "PATH_INFO=%s",
	      thd->cgi.path_info);

    if (thd->cgi.path_info[1] == '~') {
      String	name;
      size_t	len = strcspn(thd->cgi.path_info + TILDE_OFFSET, "/");
      String	home;

      name = xstrndup(thd->cgi.path_info + TILDE_OFFSET, len);
      home = getHomeDir(name);
      if (home == NULL) {	/* invalid user: no home directory  */
	xasprintf(&thd->cgi.env[thd->cgi.envno++], "PATH_TRANSLATED=%s/%%7E%s",
		  httpd->vhost->docRoot, thd->cgi.path_info + TILDE_OFFSET);
      }
      else {
	xasprintf(&thd->cgi.env[thd->cgi.envno++], "PATH_TRANSLATED=%s/%s%s",
		  home, httpd->vhost->userDir,
		  thd->cgi.path_info + TILDE_OFFSET + len);
      }
      xfree(name);
    }
    else {
      xasprintf(&thd->cgi.env[thd->cgi.envno++], "PATH_TRANSLATED=%s%s",
		httpd->vhost->docRoot, thd->cgi.path_info);
    }
  }

  xasprintf(&thd->cgi.env[thd->cgi.envno++], "QUERY_STRING=%s",
	    thd->cgi.query_string == NULL ? "" : thd->cgi.query_string);

  if (thd->client.name != NULL) {
    xasprintf(&thd->cgi.env[thd->cgi.envno++], "REMOTE_HOST=%s",
	      thd->client.name);
  }

  xasprintf(&thd->cgi.env[thd->cgi.envno++], "REMOTE_ADDR=%s", thd->client.ip);
  xasprintf(&thd->cgi.env[thd->cgi.envno++], "REMOTE_PORT=%d",
	    thd->client.port);
  xasprintf(&thd->cgi.env[thd->cgi.envno++], "AUTH_TYPE=Basic");

  if (thd->authUser != NULL) {
    xasprintf(&thd->cgi.env[thd->cgi.envno++], "REMOTE_USER=%s", thd->authUser);
  }

  if (thd->identUser != NULL) {
    xasprintf(&thd->cgi.env[thd->cgi.envno++], "REMOTE_IDENT=%s",
	      thd->identUser);
  }

  if (thd->method.type == M_POST) {
    if (thd->request.content_type != NULL) {
      xasprintf(&thd->cgi.env[thd->cgi.envno++], "CONTENT_TYPE=%s",
		thd->request.content_type);
    }
    if (thd->request.content_length != 0) {
      xasprintf(&thd->cgi.env[thd->cgi.envno++], "CONTENT_LENGTH=%d",
		thd->request.content_length);
    }
  }

  /* no standard environment variable */
  if (thd->request.accept != NULL) {
#if	NO_DEBUG
    fprintf(stderr, "[env]: HTTP_ACCEPT=%s\n", thd->request.accept);
#endif	/* DEBUG */
    xasprintf(&thd->cgi.env[thd->cgi.envno++], "HTTP_ACCEPT=%s",
	      thd->request.accept);
  }

  if (thd->request.accept_encoding != NULL) {
    xasprintf(&thd->cgi.env[thd->cgi.envno++], "HTTP_ACCEPT_ENCODING=%s",
	      thd->request.accept_encoding);
  }

  if (thd->request.accept_language != NULL) {
    xasprintf(&thd->cgi.env[thd->cgi.envno++], "HTTP_ACCEPT_LANGUAGE=%s",
	      thd->request.accept_language);
  }

  if (thd->request.user_agent != NULL) {
    xasprintf(&thd->cgi.env[thd->cgi.envno++], "HTTP_USER_AGENT=%s",
	      thd->request.user_agent);
  }

  if (thd->request.referer != NULL) {
    xasprintf(&thd->cgi.env[thd->cgi.envno++], "HTTP_REFERER=%s",
	      thd->request.referer);
  }

  if (thd->request.cookie != NULL) {
    xasprintf(&thd->cgi.env[thd->cgi.envno++], "HTTP_COOKIE=%s",
	      thd->request.cookie);
  }

  thd->cgi.env[thd->cgi.envno] = NULL;

#if	NO_DEBUG
  {
    int	i;
    for (i = 0; i < thd->envno; i++) {
      fprintf(stderr, "%s\n", thd->cgienv[i]);
    }
  }
#endif	/* DEBUG */

  return TRUE;
}


/*****************************************************************************
 *****************************************************************************/

#if	DEBUG
/*
 * Function Name:	debugServerEnvironment
 * Description:		debug server environment
 * Arguments:		none
 * Return Value:	none
 */
void
debugServerEnvironment(void)
{
  int	i;

  for (i = 0; i < httpd->envno; i++) {
    fprintf(stderr, "%s\n", httpd->env[i]);
  }
}
#endif	/* DEBUG */

