/*****************************************************************************
 *
 * FILE:	user.c
 * DESCRIPTION:	management user information
 * DATE:	Sat, Sep 18 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: user.c,v 1.24.2.3 2002/01/30 13:26:33 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void		freeUserEntry(UserInfo *);
static void		insertUserCache(UserInfo *, UserInfo *);
static UserInfo *	removeUserCache(UserInfo *);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
  				/* management user information */
static UserInfo	userTable[USER_TABLE_SIZE];
static UserInfo	userList;	/* header of user_table	*/


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	getHomeDir
 * Description:		return the user's home directory
 *			(user record is registered into cache)
 * Arguments:		user name
 * Return Value:	user's home directory
 */
String
getHomeDir(username)
	String	username;
{
  register UserInfo *	up;
  struct passwd	*	pwd;	/* pointer to password entry 	*/

  /* get user information */
  pwd = getpwnam(username);
  if (pwd == NULL) {
    return NULL;
  }

  /* check userList */
  for (up = userList.next; up != &userList; up = up->next) {
    if (up->name == NULL || up->home == NULL) {
      continue;
    }
    if (strcmp(username, up->name) == 0) {	/* found */
      removeUserCache(up);
      insertUserCache(&userList, up);
      return up->home;
    }
  }

  /* get top entry from list, and move one to end of list */
  up = removeUserCache(userList.next);
  insertUserCache(&userList, up);

  /* free old memory area */
  if (up->name != NULL) {
    freeUserEntry(up);
  }

  /* register new entry */
  xfree((String)up->name);
  up->name = xstrdup(pwd->pw_name);
  xfree((String)up->home);
  up->home = xstrdup(pwd->pw_dir);

  return up->home;
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	initUserCache
 * Description:		initialize userTable called at only once
 * Arguments:		none
 * Return Value:	none
 */
void
initUserCache(void)
{
  register int	i;

  userList.next = userList.prev = &userList;
  for (i = 0; i < USER_TABLE_SIZE; i++) {
    userTable[i].name = NULL;
    userTable[i].home = NULL;
    insertUserCache(&userList, &userTable[i]);
  }
}

#if	PARANOIA		/* for free-on-exit */
/*
 * Function Name:	freeUserCache
 * Description:		free memory area of userTable
 * Arguments:		none
 * Return Value:	none
 */
void
freeUserCache(void)
{
  register int	i;

  for (i = USER_TABLE_SIZE - 1; i >= 0; i--) {
    freeUserEntry(&userTable[i]);
  }
}
#endif	/* PARANOIA */

/*
 * Function Name:	freeUserEntry
 * Description:		free memory area of userTable
 * Arguments:		user entry
 * Return Value:	none
 */
static void
freeUserEntry(p)
	register UserInfo *	p;
{
  XFree(p->name);
  XFree(p->home);
}

/*
 * Function Name:	insertUserCache
 * Description:		insert 'new' before 'pos'
 * Arguments:		position, new slot
 * Return Value:	none
 */
static void
insertUserCache(pos, new)
	register UserInfo *	pos;
	register UserInfo *	new;
{
  pos->prev->next = new;
  new->prev	  = pos->prev;
  new->next	  = pos;
  pos->prev	  = new;
}

/*
 * Function Name:	removeUserCache
 * Description:		remove the slot from userList
 * Arguments:		removed slot
 * Return Value:	removed slot
 */
static UserInfo *
removeUserCache(x)
	register UserInfo *	x;
{
  x->prev->next = x->next;
  x->next->prev = x->prev;
  return x;
}

