/*****************************************************************************
 *
 * FILE:	wyvern.c
 * DESCRIPTION:	wyvern main source file
 * DATE:	Sun, Aug 29 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: wyvern.c,v 1.64.2.1 2002/01/30 13:26:33 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#if	SIGTSTP
#include <sys/file.h>
#include <sys/ioctl.h>
#endif	/* SIGTSTP */
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#if	HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif	/* HAVE_SYS_WAIT_H */

#define	MAIN_MODULE
#include "wyvern.h"
#undef	MAIN_MODULE
#include "functions.h"


/*****************************************************************************
 *
 *	Program & Copyright
 *
 *****************************************************************************/
const char	Program[]   = "wyvern";
const char	Copyright[] = "Copyright (C) 1999-2002 Kouichi ABE (WALL)";


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	initServerEntity(String);
#if	PARANOIA
static void	freeServerEntity(void);
#endif	/* PARANOIA */
static void	checkDirectories(void);
static void	version(void);
static void	usage(void);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/


/*
 * Function Name:	quit
 * Description:		stop the service and exit
 * Arguments:		none
 * Return Value:	none
 */
void
quit(void)
{
  sigset_t	set;

  /* all signals are blocked */
  sigfillset(&set);
  sigprocmask(SIG_BLOCK, &set, NULL);

#if	USE_SSL
  /* cleanup SSL session */
  sslCleanup();
#endif	/* USE_SSL */

  /* close service channel */
  shutdown(httpd->socket.soc, SHUT_RDWR);/* shutdown full-duplex connection */
  xclose(httpd->socket.soc);	/* close the main socket */

#if	MAIL_ANNOUNCE
  sendMail();		/* report a mail to the admin */
#endif	/* MAIL_ANNOUNCE */

  /* close all log files */
  closeLogFiles();

#if	PARANOIA		/* for free-on-exit */
  cleanupModules();		/* unload all loaded modules */
  freeMimeTypeTable();
  freeErrorDocumentTable();
  freeThreadQueue();		/* close all connected socket in each thread */
  freeServerEnvironment();	/* free memory area (env) */
  freeUserCache();		/* free memory area (userTable) */
  freeServerEntity();		/* unlink pid file & free memory area (httpd) */
  cleanupMemoryBlock();		/* cleanup all memories by memory manager */
#else
  if (httpd->pid > 0) {		/* remove PID file to unlock */
    (void)unlink(httpd->pidFile);
  }
#endif	/* PARANOIA */
#if	USE_SYSLOG
  closelog();
#endif	/* USE_SYSLOG */

  exit(0);
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	initServerEntity
 * Description:		initialize server entity
 * Arguments:		path of the configuration file
 * Return Value:	none
 */
static void
initServerEntity(configFile)
	String	configFile;
{
  httpd = (Server *)xcalloc(1, sizeof(Server));
  if (httpd == NULL) {
    fprintf(stderr, "*** WYVERN COULD NOT GET ENOUGH MEMORY ***\n");
    exit(1);
  }

  memset(&httpd->socket, 0, sizeof(httpd->socket));	/* Socket	*/
  httpd->port		= DEF_PORT;		/* Port		*/
#if	USE_SSL
  httpd->portSSL	= DEF_SECURE_PORT;	/* SecurePort	*/
#endif	/* USE_SSL */
  httpd->maxRequest	= DEF_MAX_REQUEST;	/* MaxQueue	*/
  httpd->uid		= UID_NOBODY;		/* User		*/
  httpd->gid		= UID_NOGROUP;		/* Group	*/
  httpd->pid		= 0;			/* Pid		*/
  httpd->pidFile	= NULL;			/* PidFile	*/
  httpd->errorLog	= stderr;		/* ErrorLog	*/
  httpd->accessLog	= stderr;		/* AccessLog	*/
  httpd->serverRoot	= NULL;			/* ServerRoot	*/
  httpd->timeout	= DEF_TIMEOUT;		/* Timeout	*/
  httpd->ka_timeout	= DEF_KA_TIMEOUT;	/* KeepAliveTimeout */
  httpd->ka_max		= DEF_KA_MAX;		/* KeepAliveMax	*/
  httpd->bindAddress	= NULL;			/* BindAddress	*/
  httpd->vhostList	= newVirtualHost(NULL);	/* init VirtualHost queue */
  httpd->vhost		= httpd->vhostList;	/* current virtual host */
  httpd->host		= httpd->vhostList;	/* original server host */
  httpd->cacheDir	= xstrdup(DEF_CACHE_DIR);	/* CacheDir	*/
  httpd->tmpDir		= xstrdup(DEF_TMP_DIR);	/* tmporary working directory */
  httpd->reboot		= FALSE;	/* flag whether server will reboot */

  /* preserve config (wyvern.conf) path */
  if (configFile == NULL) {
    httpd->configFile = xstrdup(DEF_CONFIG);	/* constant variable */
  }
  else {
    if (configFile[0] != '/') {
      die("specify an absolute path of wyvern.conf\n");
    }
    httpd->configFile = xstrdup(configFile);	/* constant variable */
  }
}

#if	PARANOIA		/* for free-on-exit */
/*
 * Function Name:	freeServerEntity
 * Description:		release memory area of server entity
 * Arguments:		none
 * Return Value:	none
 */
static void
freeServerEntity(void)
{
  if (httpd->pid > 0) {
    (void)unlink(httpd->pidFile);	/* remove PID file to unlock */
  }
  XFree(httpd);
}
#endif	/* PARANOIA */


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	checkDirectories
 * Description:		confirm existance of LogDir/CacheDir/TmpDir directories
 * Arguments:		none
 * Return Value:	none
 */
static void
checkDirectories(void)
{
  struct stat	sbuf;

  /* check a cache directory */
  if (lstat(httpd->cacheDir, &sbuf) == -1 && errno == ENOENT) {
    die("No cache directory: %s\n", httpd->cacheDir);
  }
  (void)lchown(httpd->cacheDir, httpd->uid, httpd->gid);/* ignore error */

  /* check a temporary directory */
  if (lstat(httpd->tmpDir, &sbuf) == -1 && errno == ENOENT) {
    die("No temporary directory: %s\n", httpd->tmpDir);
  }
  (void)lchown(httpd->tmpDir, httpd->uid, httpd->gid);	/* ignore error */
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	version
 * Description:		show the version of Wyvern
 * Arguments:		none
 * Return Value:	none
 */
static void
version(void)
{
  fprintf(stderr, "%s\n%s\n", SERVER_SOFTWARE, Copyright);
  fprintf(stderr, "[compiled at %s on %s by %s]\n",
		  __TIME__, __DATE__, COMPILED_USER);
#if	DEBUG
  fputs("+DEBUG ", stderr);
#endif	/* DEBUG */
#if	MAIL_ANNOUNCE
  fputs("+MAIL ", stderr);
#endif	/* MAIL_ANNOUNCE */
#if	USE_JP_MESG
  fputs("+JMSG ", stderr);
#endif	/* USE_JP_MESG */
#if	USE_IPV6
  fputs("+IPv6 ", stderr);
#endif	/* USE_IPV6 */
#if	USE_SYSLOG
  fputs("+SYSLOG ", stderr);
#endif	/* USE_SYSLOG */
#ifndef	USE_CONTENT_MD5
  fputs("-MD5 ", stderr);
#endif	/* USE_CONTENT_MD5 */
#if	BUGGY_KEEP_ALIVE
  fputs("+KEEP ", stderr);
#endif	/* BUGGY_KEEP_ALIVE */
#if	USE_SSL
  fputs("+SSL ", stderr);
#endif	/* USE_SSL */
  fputs("\n\n", stderr);
  exit(2);
}

/*
 * Function Name:	usage
 * Description:		show the simple usage for Wyvern
 * Arguments:		none
 * Return Value:	none
 */
static void
usage(void)
{
  fprintf(stderr,
	  "%s\n%s\n%s\n"
	  "usage: %s [-r config_file][-p port]\n\n",
	  SERVER_SOFTWARE,
	  Copyright,
	  SERVER_ADDRESS,
	  Program);
  exit(3);
}


/*****************************************************************************
 *****************************************************************************/

/*
 *	Wyvern Main
 */
int
main(argc, argv)
	int	argc;
	char	**argv;
{
#if	defined(__sun__)
  extern char *	optarg;
  extern int	optind;
#endif	/* __sun__ */
  String	config	= NULL;		/* use default config file if NULL */
  unsigned int	port	= 0;		/* use default port if 0 */
  int		ch;			/* variables for arguments */

  /* treat arguments */
  while ((ch = getopt(argc, argv, "r:p:vh?")) != -1) {
    switch (ch) {
      case 'r':
	config = optarg;
	break;
      case 'p':
	port = atoi(optarg);
	break;
      case 'v':
	version();
      case 'h':
      case '?':
      default:
	usage();
    }
  }
  argc -= optind;
  argv += optind;

  /*
   * bootstrap
   */
#ifndef	DEBUG
  detach();			/* detach the control terminal */
#endif	/* DEBUG */

  initServerEntity(config);	/* create server entity */
  initPluggableModules();	/* initialize module interface */
  initErrorDocumentTable();	/* initialize error document table */
  initMimeTypeTable();		/* initialize mime-type table */

  readConfigFile();		/* read configuration file */

  checkDirectories();		/* check CacheDir/TmpDir directories */

  openServiceChannel(port);	/* open service channel (socket) */

#if	USE_SSL
  sslOpenServiceChannel();	/* initialize SSL mode (ignore error) */
#endif	/* USE_SSL */

#if	USE_SYSLOG
  openlog(Program, LOG_PID, LOG_DAEMON);
#endif	/* USE_SYSLOG */
  openLogFiles();		/* initialize the log files */

#ifndef	DEBUG
  storePID();			/* preserve server process ID	*/
#endif	/* DEBUG */

  setSystemResource();		/* limit system resources */
  setOwner();			/* change owner/group of process */

  initThreadQueue();		/* initialize the thread queue */
  initServerEnvironment();	/* initialize the server environment */
  initUserCache();		/* initialize the user cache list */

  setSignalTrap();		/* set signals */

#if	NO_DEBUG
  debugServerEnvironment();
#endif	/* DEBUG */

  /*
   * main-loop
   */
  foreverRunning();

  return 0;	/* Not reached!	*/
}

